#define _POSIX_C_SOURCE 200112L
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <sys/mman.h>
#include <semaphore.h>
#include <stdlib.h>
#include <sys/stat.h>

#define DATA_AVAILABLE	 (0)
#define DATA_UNAVAILABLE (1)
#define DATA_COMPLETE    (2)

#define MBOX_SIZE	(1024)

/** struktura reprezentujici schranku, do ktere je mozne ulozit nanejvys jednu zpravu */
struct mbox {
	sem_t lock;           // sdileny semafor
	char status;          // stav schranky (viz makra DATA_*)
	char data[MBOX_SIZE]; // samotna data
};

/** inicializuje obsah schranky */
void mbox_init(struct mbox *mbox)
{
	sem_init(&(mbox->lock), 1, 1);
	mbox->status = DATA_UNAVAILABLE;
}

/**
 * Pokud je schranka prazdna, ulozi do ni zpravu v podobe celeho sicla.
 * V opacnem pripade ceka na uvolneni schranky.
 */
void mbox_send(struct mbox *mbox, int value)
{
	int send = 0;
	while (!send) {
		sem_wait(&mbox->lock);
		if (mbox->status == DATA_UNAVAILABLE) {
			sprintf(mbox->data, "msg: %i", value);
			mbox->status = DATA_AVAILABLE;
			printf("SENDER: %s\n", mbox->data);
			send = 1;
		}
		sem_post(&mbox->lock);
		//sleep(1);
	}
}

/**
 * Ze schranky vyzvedna zpravu a vypise ji na std. vystup.
 * Pokud zprava ve schrance neni, ceka.
 * Navratovou hodnotou je signalizovano uspesne precteni dat,
 * nebo konec komunikace.
 */
int mbox_receive(struct mbox *mbox)
{
	int read = 0; // signalizuje uspesne precteni dat
	int prev_status = 0;
	while (!read) {
		sem_wait(&mbox->lock);
		prev_status = mbox->status;
		switch (mbox->status) {
			case DATA_AVAILABLE: 
				printf("%s\n", mbox->data);
			case DATA_COMPLETE:
				mbox->status = DATA_UNAVAILABLE;
				read = 1;
				break;
		}
		sem_post(&mbox->lock);
	}
	return prev_status;
}

void mbox_destroy(struct mbox *mbox)
{
	// FIXME:
	int success = 0;
	// signalizuje konec komunikace
	while (!success) {
		sem_wait(&mbox->lock);
		if (mbox->status == DATA_UNAVAILABLE) {
			mbox->status = DATA_COMPLETE;
			success = 1;
		}
		sem_post(&mbox->lock);
	}

	// ceka, az prijemce zpravu o konci komunikace obdrzi
	while (!success) {
		sem_wait(&mbox->lock);
		if (mbox->status == DATA_UNAVAILABLE) {
			success = 1;
		}
		sem_post(&mbox->lock);
	}

	sem_destroy(&mbox->lock);
}


static void show_help(char *app_name)
{
	printf("Usage: %s [writer|reader]\n", app_name);
}

int main(int argc, char **argv)
{
	if (argc < 2) {
		show_help(argv[0]);
		return 0;
	}

	// otevre/namapuje do pameti soubor pro sdileni dat
	size_t length = sizeof(struct mbox);
	int fd = open("shm.dat", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	//int fd = shm_open("my_shm_file", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	ftruncate(fd, length);

	struct mbox *mb = mmap(NULL, length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	printf("Buffer at:%p\n", mb);
	
	if (!strcmp(argv[1], "writer")) {
		// inicializuje mbfer
		mbox_init(mb);
		// zasila cisla z interval 0 ... 99
		for (int i = 0; i < 100; i++) {
			mbox_send(mb, i);
			//sleep(1);
		}
		// odstrani mbfer
		mbox_destroy(mb);
	} else if (!strcmp(argv[1], "reader")) {
		int status;
		do {
			// prijima zpravy
			status = mbox_receive(mb);
		} while (status != DATA_COMPLETE);
	} else {
		show_help(argv[0]);
	}

	
	munmap(mb, length);
	close(fd);
	//shm_unlink("my_shm_file");
	return 0;
}
